﻿using System;
using System.Collections.Generic;
using System.IdentityModel.Policy;
using System.Linq;
using System.Security.Cryptography;
using System.Security.Principal;
using System.ServiceModel;
using System.ServiceModel.Channels;
using System.ServiceModel.Security;
using System.Text;

namespace WcfRestServiceDemo.Service
{
    public class AuthorizationManager : ServiceAuthorizationManager
    {
        private const string AuthHeaderName = "Authorization";
        private const string OAuthPrefix = "OAuth ";
        private const string OAuthConsumerKey = "oauth_consumer_key";
        private const string OAuthNonce = "oauth_nonce";
        private const string OAuthSignatureMethod = "oauth_signature_method";
        private const string OAuthSignature = "oauth_signature";
        private const string OAuthTimestamp = "oauth_timestamp";
        private const string OAuthToken = "oauth_token";
        private const string OAuthVersion = "oauth_version";

        protected override bool CheckAccessCore(OperationContext operationContext)
        {
            var authParamString = operationContext.GetHttpHeader(AuthHeaderName);
            if(authParamString == null || !authParamString.StartsWith(OAuthPrefix))
            {
                SetAnonymousContext(operationContext.RequestContext.RequestMessage);
                return true;
            }
            var authParams = authParamString.Substring(OAuthPrefix.Length).Split(',')
                .Select(p => p.Split('=')).ToDictionary(pair => pair[0], pair => pair[1].Trim('"'));
            var uri = operationContext.RequestContext.RequestMessage.Headers.To.AbsoluteUri;
            var method = ((HttpRequestMessageProperty)
                 operationContext.RequestContext.RequestMessage.Properties[HttpRequestMessageProperty.Name]).Method;

            // Construct the base string.
            var baseString = JoinParameters(method, uri,
                string.Format("{0}={1}&{2}={3}&{4}={5}&{6}={7}&{8}={9}&{10}={11}",
                OAuthConsumerKey, authParams[OAuthConsumerKey],
                OAuthNonce, authParams[OAuthNonce],
                OAuthSignatureMethod, authParams[OAuthSignatureMethod],
                OAuthTimestamp, authParams[OAuthTimestamp],
                OAuthToken, authParams[OAuthToken],
                OAuthVersion, authParams[OAuthVersion]));

            var signature = GetSignature(baseString, authParams[OAuthConsumerKey]);
            if(Uri.EscapeDataString(signature) != authParams[OAuthSignature])
            {
                SetAnonymousContext(operationContext.RequestContext.RequestMessage);
                return true;
            }

            // Authenticated successfully.
            SetSecurityContext(operationContext.RequestContext.RequestMessage, authParams[OAuthConsumerKey]);
            return true;
        }

        private string GetSignature(string baseString, string key)
        {
            var hmacsha1 = new HMACSHA1(Encoding.UTF8.GetBytes(GetAuthSecret(key) + "&"));
            return Convert.ToBase64String(hmacsha1.ComputeHash(Encoding.UTF8.GetBytes(baseString)));
        }

        private static string JoinParameters(params string[] parameters)
        {
            return string.Join("&", parameters.Select(Uri.EscapeDataString));
        }

        private static string GetAuthSecret(string authKey)
        {
            // TODO: Get actual secret for user.
            return "kd94hf93k423kf44";
        }

        private static void SetAnonymousContext(Message request)
        {
            SetSecurityContext(request, string.Empty);
        }

        private static void SetSecurityContext(Message request, string username)
        {
            var policies = new List<IAuthorizationPolicy> { new AuthorizationPolicy(new GenericIdentity(username)) };
            var securityContext = new ServiceSecurityContext(policies.AsReadOnly());

            if(request.Properties.Security != null)
            {
                request.Properties.Security.ServiceSecurityContext = securityContext;
            }
            else
            {
                request.Properties.Security = new SecurityMessageProperty { ServiceSecurityContext = securityContext };
            }
        }
    }
}